from rlbot.utils.game_state_util import GameState, BallState, CarState, GameInfoState, Physics, Vector3, Rotator


def dict_to_game_state(state_dict):
    gs = GameState()
    if 'ball' in state_dict:
        gs.ball = BallState()
        if 'physics' in state_dict['ball']:
            gs.ball.physics = dict_to_physics(state_dict['ball']['physics'])
    if 'cars' in state_dict:
        gs.cars = {}
        for index, car in state_dict['cars'].items():
            car_state = CarState()
            if 'physics' in car:
                car_state.physics = dict_to_physics(car['physics'])
            if 'boost_amount' in car:
                car_state.boost_amount = car['boost_amount']
            gs.cars[int(index)] = car_state
    if 'game_info' in state_dict:
        gs.game_info = GameInfoState()
        if 'paused' in state_dict['game_info']:
            gs.game_info.paused = state_dict['game_info']['paused']
        if 'world_gravity_z' in state_dict['game_info']:
            gs.game_info.world_gravity_z = state_dict['game_info']['world_gravity_z']
        if 'game_speed' in state_dict['game_info']:
            gs.game_info.game_speed = state_dict['game_info']['game_speed']
    if 'console_commands' in state_dict:
        gs.console_commands = state_dict['console_commands']
    return gs


def dict_to_physics(physics_dict):
    phys = Physics()
    if 'location' in physics_dict:
        phys.location = dict_to_vec(physics_dict['location'])
    if 'velocity' in physics_dict:
        phys.velocity = dict_to_vec(physics_dict['velocity'])
    if 'angular_velocity' in physics_dict:
        phys.angular_velocity = dict_to_vec(physics_dict['angular_velocity'])
    if 'rotation' in physics_dict:
        phys.rotation = dict_to_rot(physics_dict['rotation'])
    return phys


def dict_to_vec(v):
    vec = Vector3()
    if 'x' in v:
        vec.x = v['x']
    if 'y' in v:
        vec.y = v['y']
    if 'z' in v:
        vec.z = v['z']
    return vec


def dict_to_rot(r):
    rot = Rotator()
    if 'pitch' in r:
        rot.pitch = r['pitch']
    if 'yaw' in r:
        rot.yaw = r['yaw']
    if 'roll' in r:
        rot.roll = r['roll']
    return rot



import os
import shutil
import subprocess
import sys
import json
from pathlib import Path

import eel
from PyQt5.QtCore import QSettings
from PyQt5.QtWidgets import QApplication, QFileDialog
from rlbot.parsing.agent_config_parser import create_looks_configurations, BOT_CONFIG_LOADOUT_HEADER, \
    BOT_CONFIG_LOADOUT_ORANGE_HEADER, BOT_CONFIG_LOADOUT_PAINT_BLUE_HEADER, BOT_CONFIG_LOADOUT_PAINT_ORANGE_HEADER, \
    load_bot_appearance
from rlbot.parsing.bot_config_bundle import get_bot_config_bundle, get_script_config_bundle, RunnableConfigBundle, \
    BotConfigBundle
from rlbot.parsing.directory_scanner import scan_directory_for_bot_configs, scan_directory_for_script_configs
from rlbot.parsing.match_settings_config_parser import map_types, game_mode_types, \
    boost_amount_mutator_types, match_length_types, max_score_types, overtime_mutator_types, \
    series_length_mutator_types, game_speed_mutator_types, ball_max_speed_mutator_types, ball_type_mutator_types, \
    ball_weight_mutator_types, ball_size_mutator_types, ball_bounciness_mutator_types, rumble_mutator_types, \
    boost_strength_mutator_types, gravity_mutator_types, demolish_mutator_types, respawn_time_mutator_types, \
    existing_match_behavior_types
from rlbot.utils.requirements_management import install_requirements_file

from rlbot_gui.bot_management.bot_creation import bootstrap_python_bot, bootstrap_scratch_bot, \
    bootstrap_python_hivemind, convert_to_filename
from rlbot_gui.bot_management.downloader import BotpackStatus, RepoDownloader, BotpackUpdater, get_json_from_url, \
    MapPackUpdater
from rlbot_gui.match_runner.match_runner import hot_reload_bots, shut_down, start_match_helper, \
    do_infinite_loop_content, spawn_car_in_showroom, set_game_state, fetch_game_tick_packet
from rlbot_gui.match_runner.custom_maps import find_all_custom_maps
from rlbot_gui.type_translation.packet_translation import convert_packet_to_dict
from rlbot_gui.persistence.settings import load_settings, BOT_FOLDER_SETTINGS_KEY, MATCH_SETTINGS_KEY, \
    LAUNCHER_SETTINGS_KEY, TEAM_SETTINGS_KEY, load_launcher_settings, launcher_preferences_from_map

#### LOAD JUST TO EXPOSE STORY_MODE
from rlbot_gui.story import story_runner
####

DEFAULT_BOT_FOLDER = 'default_bot_folder'
BOTPACK_FOLDER = 'RLBotPackDeletable'
MAPPACK_FOLDER = 'RLBotMapPackDeletable'
MAPPACK_REPO = ("azeemba", "RLBotMapPack")
OLD_BOTPACK_FOLDER = 'RLBotPack'
BOTPACK_REPO_OWNER = 'RLBot'
BOTPACK_REPO_NAME = 'RLBotPack'
BOTPACK_REPO_BRANCH = 'master' # can't change with the new release system
CREATED_BOTS_FOLDER = 'MyBots'
COMMIT_ID_KEY = 'latest_botpack_commit_id'
bot_folder_settings = None


@eel.expose
def start_match(bot_list, match_settings):
    launcher_preference_map = load_launcher_settings()
    launcher_prefs = launcher_preferences_from_map(launcher_preference_map)
    eel.spawn(start_match_helper, bot_list, match_settings, launcher_prefs)


@eel.expose
def kill_bots():
    shut_down()


@eel.expose
def pick_bot_folder():
    filename = pick_location(True)

    if filename:
        global bot_folder_settings
        bot_folder_settings['folders'][filename] = {'visible': True}
        settings = load_settings()
        settings.setValue(DEFAULT_BOT_FOLDER, filename)
        settings.setValue(BOT_FOLDER_SETTINGS_KEY, bot_folder_settings)
        settings.sync()
        return scan_for_bots()

    return []


def serialize_bundle(bundle: BotConfigBundle):
    return {
        'name': bundle.name,
        'type': 'rlbot',
        'skill': 1,
        'image': 'imgs/rlbot.png',
        'path': bundle.config_path,
        'looks_path': bundle.looks_path,
        'info': read_info(bundle),
        'logo': try_copy_logo(bundle),
        'missing_python_packages': [r.line for r in bundle.get_missing_python_packages() + bundle.get_python_packages_needing_upgrade()],
    }


def serialize_script_bundle(bundle):
    return {
        'name': bundle.name,
        'type': 'script',
        'image': 'imgs/rlbot.png',
        'path': bundle.config_path,
        'info': read_info(bundle),
        'logo': try_copy_logo(bundle),
        'missing_python_packages': [r.line for r in bundle.get_missing_python_packages() + bundle.get_python_packages_needing_upgrade()],
    }


def load_bot_bundle(filename):
    try:
        bundle = get_bot_config_bundle(filename)
        return [serialize_bundle(bundle)]
    except Exception as e:
        print(e)

    return []


def load_script_bundle(filename):
    try:
        bundle = get_script_config_bundle(filename)
        return [serialize_script_bundle(bundle)]
    except Exception as e:
        print(e)

    return []


@eel.expose
def pick_bot_config():
    filename = pick_location(False)
    bundle = load_script_bundle(filename) or load_bot_bundle(filename)

    if bundle:
        bot_folder_settings["files"][filename] = {"visible": True}
        settings = QSettings("rlbotgui", "preferences")
        settings.setValue(BOT_FOLDER_SETTINGS_KEY, bot_folder_settings)
        settings.sync()

    return bundle


@eel.expose
def get_folder_settings():
    return bot_folder_settings


@eel.expose
def save_folder_settings(folder_settings):
    global bot_folder_settings
    bot_folder_settings = folder_settings
    settings = QSettings('rlbotgui', 'preferences')
    settings.setValue(BOT_FOLDER_SETTINGS_KEY, bot_folder_settings)
    settings.sync()


def validate_bots(bots):
    '''Reload rlbot controlled bot bundles and remove invalid ones'''
    valid_bots = []

    for bot in bots:
        if bot["type"] in ('rlbot', 'party_member_bot'):
            valid_bots += load_bot_bundle(bot["path"])
        else:
            valid_bots.append(bot)

    return valid_bots


@eel.expose
def get_match_settings():
    settings = load_settings()
    match_settings = settings.value(MATCH_SETTINGS_KEY, type=dict)
    return match_settings if match_settings else None


@eel.expose
def get_launcher_settings():
    return load_launcher_settings()


@eel.expose
def get_team_settings():
    settings = load_settings()
    team_settings = settings.value(TEAM_SETTINGS_KEY, type=dict)
    if not team_settings:
        return None

    return {
        "blue_team": validate_bots(team_settings["blue_team"]),
        "orange_team": validate_bots(team_settings["orange_team"])
    }


@eel.expose
def save_match_settings(match_settings):
    settings = load_settings()
    settings.setValue(MATCH_SETTINGS_KEY, match_settings)


@eel.expose
def save_launcher_settings(launcher_settings_map):
    settings = load_settings()
    settings.setValue(LAUNCHER_SETTINGS_KEY, launcher_settings_map)


@eel.expose
def save_team_settings(blue_bots, orange_bots):
    settings = load_settings()
    settings.setValue(TEAM_SETTINGS_KEY, {"blue_team": blue_bots, "orange_team": orange_bots})


@eel.expose
def pick_location(is_folder, filter="Config files (*.cfg)"):
    """
    We're using python for file picking because only python (not javascript) can retrieve
    an actual path on the file system which is what we need.
    https://stackoverflow.com/questions/2809688/directory-chooser-in-html-page

    :return:
    """

    app = QApplication([])
    options = QFileDialog.Options()

    if is_folder:
        filename = QFileDialog.getExistingDirectory(options=options)
    else:
        filename, _ = QFileDialog.getOpenFileName(filter=filter, options=options)

    app.exit()

    return filename


def read_info(bundle: RunnableConfigBundle):
    details_header = 'Details'
    if bundle.base_agent_config.has_section(details_header):
        return {
            'developer': bundle.base_agent_config.get(details_header, 'developer'),
            'description': bundle.base_agent_config.get(details_header, 'description'),
            'fun_fact': bundle.base_agent_config.get(details_header, 'fun_fact'),
            'github': bundle.base_agent_config.get(details_header, 'github'),
            'language': bundle.base_agent_config.get(details_header, 'language'),
        }
    return None


@eel.expose
def get_looks(path: str) -> dict:
    looks_config = create_looks_configurations().parse_file(path)
    looks = {'blue': {}, 'orange': {}}

    def serialize_category(target: dict, header_name: str):
        header = looks_config.get_header(header_name)
        for key in header.values.keys():
            if header.get(key) is not None:
                target[key] = str(header.get(key))

    serialize_category(looks['blue'], BOT_CONFIG_LOADOUT_HEADER)
    serialize_category(looks['orange'], BOT_CONFIG_LOADOUT_ORANGE_HEADER)
    serialize_category(looks['blue'], BOT_CONFIG_LOADOUT_PAINT_BLUE_HEADER)
    serialize_category(looks['orange'], BOT_CONFIG_LOADOUT_PAINT_ORANGE_HEADER)

    return looks


def convert_to_looks_config(looks: dict):
    looks_config = create_looks_configurations()

    def deserialize_category(source: dict, header_name: str):
        header = looks_config.get_header(header_name)
        for key in header.values.keys():
            if key in source:
                header.set_value(key, source[key])

    deserialize_category(looks['blue'], BOT_CONFIG_LOADOUT_HEADER)
    deserialize_category(looks['orange'], BOT_CONFIG_LOADOUT_ORANGE_HEADER)
    deserialize_category(looks['blue'], BOT_CONFIG_LOADOUT_PAINT_BLUE_HEADER)
    deserialize_category(looks['orange'], BOT_CONFIG_LOADOUT_PAINT_ORANGE_HEADER)

    return looks_config


@eel.expose
def save_looks(looks: dict, path: str):
    looks_config = convert_to_looks_config(looks)

    with open(path, 'w', encoding='utf8') as f:
        f.write(str(looks_config))
        print(f'Saved appearance to {path}')


@eel.expose
def spawn_car_for_viewing(looks: dict, team: int, showcase_type: str, map_name: str):
    looks_config = convert_to_looks_config(looks)
    loadout_config = load_bot_appearance(looks_config, team)
    launcher_settings_map = get_launcher_settings()
    launcher_prefs = launcher_preferences_from_map(launcher_settings_map)
    spawn_car_in_showroom(loadout_config, team, showcase_type, map_name, launcher_prefs)


@eel.expose
def scan_for_bots():
    bot_hash = {}

    for folder, props in bot_folder_settings['folders'].items():
        if props['visible']:
            bots = get_bots_from_directory(folder)
            for bot in bots:
                bot_hash[bot['path']] = bot

    for file, props in bot_folder_settings['files'].items():
        if props['visible']:
            bots = load_bot_bundle(file)  # Returns a list of size 1
            for bot in bots:
                bot_hash[bot['path']] = bot

    return list(bot_hash.values())


@eel.expose
def scan_for_scripts():
    script_hash = {}

    for folder, props in bot_folder_settings['folders'].items():
        if props['visible']:
            bots = get_scripts_from_directory(folder)
            for bot in bots:
                script_hash[bot['path']] = bot

    for file, props in bot_folder_settings['files'].items():
        if props['visible']:
            bots = load_script_bundle(file)  # Returns a list of size 1
            for bot in bots:
                script_hash[bot['path']] = bot

    return list(script_hash.values())


def try_copy_logo(bundle: RunnableConfigBundle):
    logo_path = bundle.get_logo_file()
    if logo_path is not None and os.path.exists(logo_path):
        web_url = 'imgs/logos/' + convert_to_filename(bundle.name) + '/' + convert_to_filename(logo_path)
        target_file = os.path.join(os.path.dirname(__file__), 'gui', web_url)
        os.makedirs(os.path.dirname(target_file), exist_ok=True)
        shutil.copy(logo_path, target_file)
        return web_url
    return None


def get_bots_from_directory(bot_directory):
    bundles = filter_hidden_bundles(scan_directory_for_bot_configs(bot_directory))
    return [serialize_bundle(bundle) for bundle in bundles]


def get_scripts_from_directory(bot_directory):
    bundles = filter_hidden_bundles(scan_directory_for_script_configs(bot_directory))
    return [serialize_script_bundle(bundle) for bundle in bundles]


def filter_hidden_bundles(bundles):
    return [bundle for bundle in bundles if not bundle.config_file_name.startswith("_")]


@eel.expose
def get_language_support():
    java_return_code = os.system("java -version 2> nul")
    # Only bother returning iffy languages. No point in sending 'python': True
    return {
        'java': java_return_code == 0,
        'chrome': is_chrome_installed(),  # Scratch bots need chrome to auto-run
        'fullpython': is_full_python(),
    }


@eel.expose
def get_match_options():
    return {
        'map_types': map_types + find_all_custom_maps(),
        'game_modes': game_mode_types,
        'match_behaviours': existing_match_behavior_types,
        'mutators': {
            'match_length_types': match_length_types,
            'max_score_types': max_score_types,
            'overtime_types': overtime_mutator_types,
            'series_length_types': series_length_mutator_types,
            'game_speed_types': game_speed_mutator_types,
            'ball_max_speed_types': ball_max_speed_mutator_types,
            'ball_type_types': ball_type_mutator_types,
            'ball_weight_types': ball_weight_mutator_types,
            'ball_size_types': ball_size_mutator_types,
            'ball_bounciness_types': ball_bounciness_mutator_types,
            'boost_amount_types': boost_amount_mutator_types,
            'rumble_types': rumble_mutator_types,
            'boost_strength_types': boost_strength_mutator_types,
            'gravity_types': gravity_mutator_types,
            'demolish_types': demolish_mutator_types,
            'respawn_time_types': respawn_time_mutator_types
        }
    }


@eel.expose
def install_package(package_string):
    exit_code = subprocess.call([sys.executable, "-m", "pip", "install", '--upgrade', package_string])
    print(exit_code)
    return {'exitCode': exit_code, 'package': package_string}


@eel.expose
def install_requirements(config_path):
    try:
        bundle = get_bot_config_bundle(config_path)
    except Exception:
        bundle = get_script_config_bundle(config_path)

    if bundle.requirements_file:
        exit_code = install_requirements_file(bundle.requirements_file)
        return {'exitCode': exit_code, 'package': bundle.requirements_file}
    else:
        return {'exitCode': 1, 'package': None}


def get_last_botpack_commit_id():
    url = f'https://api.github.com/repos/{BOTPACK_REPO_OWNER}/{BOTPACK_REPO_NAME}/branches/{BOTPACK_REPO_BRANCH}'
    try:
        return get_json_from_url(url)['commit']['sha']
    except:
        return 'unknown'

@eel.expose
def get_downloaded_botpack_commit_id():
    settings = load_settings()
    local_commit_id = settings.value(COMMIT_ID_KEY, type=str)
    return local_commit_id


@eel.expose
def is_botpack_up_to_date():
    local_commit_id = get_downloaded_botpack_commit_id()

    if not local_commit_id:
        return False

    github_commit_id = get_last_botpack_commit_id()

    if github_commit_id == 'unknown':
        return True

    return github_commit_id == local_commit_id


def get_content_folder():
    if Path(os.getcwd()).name == 'RLBotGUI':
        # This is the classic case where we're using embedded python, and our current working directory is already
        # something like C:\Users\tareh\AppData\Local\RLBotGUI. It will also match if you have cloned RLBotGUI.
        return Path(os.getcwd())

    # If we get here, we're likely using system python or something along those lines.
    local_app_data = os.getenv('LOCALAPPDATA')
    if local_app_data:
        # We appear to be on Windows platform. This is where the new Windows launch script puts things.
        guix_path = Path(os.getenv('LOCALAPPDATA')) / 'RLBotGUIX'
        guix_path.mkdir(exist_ok=True)
        return guix_path

    # Probably on linux or mac at this point. Go with CWD again, because that's what they've used historically.
    return Path(os.getcwd())


def update_gui_after_botpack_update(botpack_location, botpack_status, additional_settings=None):
    if botpack_status is BotpackStatus.SUCCESS:
        # Configure the folder settings.
        bot_folder_settings['folders'][str(botpack_location)] = {'visible': True}

        settings = load_settings()
        settings.setValue(BOT_FOLDER_SETTINGS_KEY, bot_folder_settings)
        if additional_settings:
            for key, value in additional_settings.items():
                settings.setValue(key, value)
        settings.sync()
        scan_for_bots()


@eel.expose
def download_bot_pack():
    botpack_location = get_content_folder() / BOTPACK_FOLDER
    botpack_status = RepoDownloader().download(BOTPACK_REPO_OWNER, BOTPACK_REPO_NAME, botpack_location)

    additional_settings = {COMMIT_ID_KEY: get_last_botpack_commit_id()}
    update_gui_after_botpack_update(botpack_location, botpack_status, additional_settings)


@eel.expose
def update_map_pack():
    location = get_content_folder() / MAPPACK_FOLDER

    updater = MapPackUpdater(location, MAPPACK_REPO[0], MAPPACK_REPO[1])
    map_index_old = updater.get_map_index()
    status = updater.needs_update()

    if status == BotpackStatus.REQUIRES_FULL_DOWNLOAD:
        status = RepoDownloader().download(
            MAPPACK_REPO[0],
            MAPPACK_REPO[1],
            location,
            update_tag_setting=False)
        
        map_index = updater.get_map_index()

        if map_index is None:
            print("ERROR: Updating mappack failed! There is no revision")
            return
        updater.hydrate_map_pack(map_index_old)

    update_gui_after_botpack_update(location, status)


@eel.expose
def get_map_pack_revision():
    location = get_content_folder() / MAPPACK_FOLDER
    updater = MapPackUpdater(location, MAPPACK_REPO[0], MAPPACK_REPO[1])
    index = updater.get_map_index()
    return index["revision"] if "revision" in index else None


@eel.expose
def update_bot_pack():
    botpack_location = get_content_folder() / BOTPACK_FOLDER
    botpack_status = BotpackUpdater().update(BOTPACK_REPO_OWNER, BOTPACK_REPO_NAME, botpack_location)

    if botpack_status == BotpackStatus.REQUIRES_FULL_DOWNLOAD:
        botpack_status = RepoDownloader().download(BOTPACK_REPO_OWNER, BOTPACK_REPO_NAME, botpack_location)

    additional_settings = {COMMIT_ID_KEY: get_last_botpack_commit_id()}
    update_gui_after_botpack_update(botpack_location, botpack_status, additional_settings)


@eel.expose
def get_recommendations():
    try:
        # Load recommendations.json
        botpack_folder = get_content_folder() / BOTPACK_FOLDER
        file = open(botpack_folder / f'{BOTPACK_REPO_NAME}-{BOTPACK_REPO_BRANCH}' / 'RLBotPack' / 'recommendations.json')
        data = json.load(file)

        # Replace bot names with bot bundles with matching names from the botpack
        bots_in_botpack = get_bots_from_directory(botpack_folder)
        for recommendation in data['recommendations']:
            bot_names = recommendation['bots']
            bots = [next(bot for bot in bots_in_botpack if bot['name'] == name) for name in bot_names]
            recommendation['bots'] = bots

        return data

    except:
        return None


@eel.expose
def show_bot_in_explorer(bot_cfg_path):
    import subprocess
    directory = os.path.dirname(bot_cfg_path)
    subprocess.Popen(f'explorer "{directory}"')


@eel.expose
def hot_reload_python_bots():
    hot_reload_bots()


def ensure_bot_directory():
    content_folder = get_content_folder()
    bot_directory = content_folder / CREATED_BOTS_FOLDER
    bot_directory.mkdir(exist_ok=True)

    bot_folder_settings['folders'][str(bot_directory)] = {'visible': True}
    settings = load_settings()
    settings.setValue(BOT_FOLDER_SETTINGS_KEY, bot_folder_settings)
    settings.sync()

    return bot_directory


@eel.expose
def begin_python_bot(bot_name):
    bot_directory = ensure_bot_directory()

    try:
        config_file = bootstrap_python_bot(bot_name, bot_directory)
        return {'bots': load_bot_bundle(config_file)}
    except FileExistsError as e:
        return {'error': str(e)}


@eel.expose
def begin_scratch_bot(bot_name):
    bot_directory = ensure_bot_directory()

    try:
        config_file = bootstrap_scratch_bot(bot_name, bot_directory)
        install_package('webdriver_manager')  # Scratch bots need this, and the GUI's python doesn't have it by default.
        return {'bots': load_bot_bundle(config_file)}
    except FileExistsError as e:
        return {'error': str(e)}


@eel.expose
def begin_python_hivemind(hive_name):
    bot_directory = ensure_bot_directory()

    try:
        config_file = bootstrap_python_hivemind(hive_name, bot_directory)
        return {'bots': load_bot_bundle(config_file)}
    except FileExistsError as e:
        return {'error': str(e)}


@eel.expose
def set_state(state):
    set_game_state(state)


@eel.expose
def fetch_game_tick_packet_json():
    """
    Here we're selectively converting parts of the game tick packet into dict format,
    which will automatically get serialized into json by eel. We're skipping over parts
    that are currently not needed by the GUI, e.g. boost pad state.
    """
    ctypes_packet = fetch_game_tick_packet()
    dict_version = convert_packet_to_dict(ctypes_packet)
    return dict_version  # This gets converted to json automatically.


should_quit = False


def on_websocket_close(page, sockets):
    global should_quit
    eel.sleep(3.0)  # We might have just refreshed. Give the websocket a moment to reconnect.
    if not len(eel._websockets):
        # At this point we think the browser window has been closed.
        should_quit = True
        shut_down()


def is_chrome_installed():
    # Lots of hasattr checks because we're currently stuck supporting multiple versions of eel at once.
    if hasattr(eel.browsers, 'chr'):
        return eel.browsers.chr.get_instance_path() is not None
    else:
        chm = eel.browsers.chm
        if hasattr(chm, 'get_instance_path'):
            return chm.get_instance_path() is not None

        return chm.find_path() is not None


def is_full_python():
    # As opposed to embedded python. A full python installation is better at package management,
    # has access to tkinter, etc. This logic might be brittle; it's based on the historical fact that
    # our embedded installer always put python within the RLBotGUI directory.
    if 'RLBotGUI' in Path(sys.executable).parts:
        return False
    return True


def launch_eel(use_chrome):
    port = 51993
    options = {'port': port}

    if use_chrome:
        # Don't put anything in the options dict. The dict is used by old eel,
        # and old eel (0.10) defaults nicely to chrome app.
        browser_mode = 'chrome'  # New eel (1.0) needs it to be 'chrome'.
    else:
        browser_mode = 'system-default'
        options['mode'] = browser_mode

    # This disable_cache thing only works if you have tare's fork of eel https://github.com/ChrisKnott/Eel/pull/102
    # installed to pip locally using this technique https://stackoverflow.com/a/49684835
    # The suppress_error=True avoids the error "'options' argument deprecated in v1.0.0", we need to keep the
    # options argument since a lot of our user base has an older version of eel.
    eel.start('main.html', size=(1300, 870), block=False, callback=on_websocket_close, options=options,
              disable_cache=True, mode=browser_mode, port=port, suppress_error=True)


def init_settings():
    settings = load_settings()
    global bot_folder_settings
    bot_folder_settings = settings.value(BOT_FOLDER_SETTINGS_KEY, type=dict)

    if not bot_folder_settings:
        bot_folder_settings = {'files': {}, 'folders': {}}
        default_folder = settings.value(DEFAULT_BOT_FOLDER, type=str)
        if default_folder:
            bot_folder_settings['folders'][default_folder] = {'visible': True}


def start():
    init_settings()
    gui_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'gui')
    eel.init(gui_folder)

    try:
        if not is_chrome_installed():
            raise Exception("Chrome does not appear to be installed.")
        launch_eel(use_chrome=True)
    except Exception as e:
        print(f'Falling back to system default browser because: {str(e)}')
        launch_eel(use_chrome=False)

    while not should_quit:
        do_infinite_loop_content()
        eel.sleep(1.0)



from PyQt5.QtCore import QSettings
from rlbot.setup_manager import DEFAULT_LAUNCHER_PREFERENCE, RocketLeagueLauncherPreference

BOT_FOLDER_SETTINGS_KEY = 'bot_folder_settings'
MATCH_SETTINGS_KEY = 'match_settings'
LAUNCHER_SETTINGS_KEY = 'launcher_settings'
TEAM_SETTINGS_KEY = 'team_settings'


def load_settings() -> QSettings:
    return QSettings('rlbotgui', 'preferences')


def launcher_preferences_from_map(launcher_preference_map: dict) -> RocketLeagueLauncherPreference:
    return RocketLeagueLauncherPreference(launcher_preference_map['preferred_launcher'],
                                          launcher_preference_map['use_login_tricks'])


def load_launcher_settings():
    settings = load_settings()
    launcher_settings = settings.value(LAUNCHER_SETTINGS_KEY, type=dict)
    return launcher_settings if launcher_settings else DEFAULT_LAUNCHER_PREFERENCE.__dict__



from datetime import datetime
from os import path
from typing import List, Tuple

import json
import random
import time
import traceback

import eel
from rlbot.matchconfig.loadout_config import LoadoutConfig
from rlbot.parsing.bot_config_bundle import get_bot_config_bundle
from rlbot.parsing.agent_config_parser import load_bot_appearance
from rlbot.utils.game_state_util import GameState, CarState
from rlbot.utils.structures.game_data_struct import GameTickPacket, Vector3
from rlbot.parsing.match_settings_config_parser import (
    game_mode_types,
    boost_amount_mutator_types,
    map_types,
    max_score_types,
    rumble_mutator_types,
    match_length_types,
)
from rlbot.matchconfig.match_config import (
    PlayerConfig,
    MatchConfig,
    MutatorConfig,
    Team,
)
from rlbot.setup_manager import SetupManager, RocketLeagueLauncherPreference

from rlbot_gui.match_runner.match_runner import get_fresh_setup_manager, setup_match


from rlbot_gui import gui as rlbot_gui  # TODO: Need to remove circular import

WITNESS_ID = random.randint(0, 1e5)
RENDERING_GROUP = "STORY"

DEBUG_MODE_SHORT_GAMES = False

def setup_failure_freeplay(setup_manager: SetupManager, message: str, color_key="red"):
    setup_manager.shut_down()
    match_config = MatchConfig()
    match_config.game_mode = game_mode_types[0]
    match_config.game_map = "BeckwithPark"
    match_config.enable_rendering = True

    mutators = MutatorConfig()
    mutators.match_length = match_length_types[3]
    match_config.mutators = mutators

    match_config.player_configs = []

    setup_manager.load_match_config(match_config)
    setup_manager.start_match()

    # wait till num players is 0
    wait_till_cars_spawned(setup_manager, 0)

    color = getattr(setup_manager.game_interface.renderer, color_key)()
    setup_manager.game_interface.renderer.begin_rendering(RENDERING_GROUP)
    # setup_manager.game_interface.renderer.draw_rect_2d(20, 20, 800, 800, True, setup_manager.game_interface.renderer.black())
    setup_manager.game_interface.renderer.draw_string_2d(20, 200, 4, 4, message, color)
    setup_manager.game_interface.renderer.end_rendering()


def make_match_config(
    challenge: dict, upgrades: dict, player_configs: List[PlayerConfig]
):
    """Setup the match, following the challenge rules and user's upgrades
    """
    match_config = MatchConfig()

    match_config.game_mode = game_mode_types[0]  # Soccar
    if challenge.get("limitations", []).count("half-field"):
        match_config.game_mode = game_mode_types[5] # Heatseeker
    match_config.game_map = challenge.get("map")
    match_config.enable_state_setting = True

    match_config.mutators = MutatorConfig()
    match_config.mutators.max_score = challenge.get("max_score")
    if DEBUG_MODE_SHORT_GAMES:
        match_config.mutators.max_score = "3 Goals"

    if challenge.get("disabledBoost"):
        match_config.mutators.boost_amount = boost_amount_mutator_types[4]  # No boost

    if "rumble" in upgrades:
        match_config.mutators.rumble = rumble_mutator_types[1]  # All rumble

    match_config.player_configs = player_configs
    return match_config


def rlbot_to_player_config(player: dict, team: Team):
    bot_path = player["path"]
    if isinstance(bot_path, list):
        bot_path = path.join(*bot_path)

    if "$RLBOTPACKROOT" in bot_path:
        for bot_folder in rlbot_gui.bot_folder_settings["folders"].keys():
            adjusted_folder = path.join(bot_folder, "RLBotPack-master")
            subbed_path = bot_path.replace("$RLBOTPACKROOT", adjusted_folder)
            if path.exists(subbed_path):
                print("it exists!")
                bot_path = subbed_path
                break

    player_config = PlayerConfig()
    player_config.bot = True
    player_config.rlbot_controlled = True
    player_config.name = player["name"]
    player_config.team = team.value
    player_config.config_path = bot_path
    config = get_bot_config_bundle(bot_path)
    loadout = load_bot_appearance(config.get_looks_config(), team.value)
    player_config.loadout_config = loadout
    return player_config


def make_human_config(team: Team):
    player_config = PlayerConfig()
    player_config.bot = False
    player_config.rlbot_controlled = False
    player_config.human_index = 0
    player_config.team = team.value
    player_config.name = ""

    return player_config


def pysonix_to_player_config(player: dict, team: Team):
    player_config = PlayerConfig()
    player_config.bot = True
    player_config.rlbot_controlled = False
    player_config.bot_skill = player["skill"] if "skill" in player else 1
    player_config.name = player["name"]
    player_config.team = team.value
    # should be able to customize "loadout_config"

    return player_config


def bot_to_player(player: dict, team: Team):
    """Helper to choose the right config for a bot"""
    if player["type"] == "psyonix":
        return pysonix_to_player_config(player, team)
    else:
        return rlbot_to_player_config(player, team)


def make_player_configs(
    challenge: dict, human_picks: List[int], team_info: dict, teammates: List[dict], all_bots
):
    player_configs = []
    player_configs.append(make_human_config(Team.BLUE))

    for i in range(challenge["humanTeamSize"] - 1):
        print(i)
        teammate = all_bots[human_picks[i]]
        config = bot_to_player(teammate, Team.BLUE)
        config.loadout_config.custom_color_id = team_info["color_secondary"]
        player_configs.append(config)

    for opponent in challenge["opponentBots"]:
        bot = bot_to_player(all_bots[opponent], Team.ORANGE)
        color = challenge.get("city_description", {}).get("color")
        if color:
            bot.loadout_config.team_color_id = color
        player_configs.append(bot)

    return player_configs


def packet_to_game_results(game_tick_packet: GameTickPacket):
    """Take the final game_tick_packet and
    returns the info related to the final game results
    """
    players = game_tick_packet.game_cars
    human_player = next(p for p in players if not p.is_bot)

    player_stats = [
        {
            "name": p.name,
            "team": p.team,
            # these are always 0, so we don't add them
            # "spawn_id": p.spawn_id,
            # "score": p.score_info.score,
            # "goals": p.score_info.goals,
            # "own_goals": p.score_info.own_goals,
            # "assists": p.score_info.assists,
            # "saves": p.score_info.saves,
            # "shots": p.score_info.shots,
            # "demolitions": p.score_info.demolitions
        }
        for p in players
        if p.name
    ]

    scores_sorted = [
        {"team_index": t.team_index, "score": t.score} for t in game_tick_packet.teams
    ]
    scores_sorted.sort(key=lambda x: x["score"], reverse=True)
    human_won = scores_sorted[0]["team_index"] == human_player.team

    return {
        "human_team": human_player.team,
        "score": scores_sorted,  # [{team_index, score}]
        "stats": player_stats,
        "human_won": human_won,
        "timestamp": datetime.now().isoformat(),
    }


def has_user_perma_failed(challenge, manual_stats):
    """
    Check if the user has perma-failed the challenge
    meaning more time in the game doesn't change the result
    """
    if "completionConditions" not in challenge:
        return False
    failed = False
    completionConditions = challenge["completionConditions"]

    if "selfDemoCount" in completionConditions:
        survived = (
            manual_stats["recievedDemos"] <= completionConditions["selfDemoCount"]
        )
        failed = failed or not survived
    return failed

def end_by_mercy(challenge, manual_stats, results):
    """Returns true if the human team is ahead by a lot
    and the other challenges have finished"""
    challenge_completed = calculate_completion(challenge, manual_stats, results)

    mercy_difference = 5
    # ignore the team, just look at the differential
    score_differential = results["score"][0]["score"] - results["score"][1]["score"]

    return score_differential >= mercy_difference and challenge_completed


def calculate_completion(challenge, manual_stats, results):
    """
    parse challenge to file completionConditions and evaluate
    each.
    All conditions are "and"
    """
    completed = results["human_won"]
    if "completionConditions" not in challenge:
        return completed

    if has_user_perma_failed(challenge, manual_stats):
        return False

    completionConditions = challenge["completionConditions"]

    if not completionConditions.get("win", True):
        # the "win" requirement is explicitly off
        completed = True

    if "scoreDifference" in completionConditions:
        # ignore the team, jsut look at the differential
        condition = completionConditions["scoreDifference"]
        difference = results["score"][0]["score"] - results["score"][1]["score"]
        completed = completed and (difference >= condition)

    if "demoAchievedCount" in completionConditions:
        achieved = (
            manual_stats["opponentRecievedDemos"]
            >= completionConditions["demoAchievedCount"]
        )
        completed = completed and achieved

    if "goalsScored" in completionConditions:
        achieved = manual_stats["humanGoalsScored"] >= completionConditions["goalsScored"]
        completed = completed and achieved

    return completed


class ManualStatsTracker:
    def __init__(self, challenge):
        self.stats = {
            "recievedDemos": 0,  # how many times the human got demo'd
            "opponentRecievedDemos": 0,  # how many times the opponents were demo'd
            "humanGoalsScored": 0,
        }

        self._challenge = challenge
        self._player_count = challenge["humanTeamSize"] + len(challenge["opponentBots"])
        self._human_team = Team.BLUE.value  # hardcoded for now
        self._human_player_index = 0

        # helper to find discrete demo events
        self._in_demo_state = [False] * self._player_count
        # helper to find who scored!
        self._last_touch_by_team = [None, None]
        self._last_score_by_team = [0, 0]

    def updateStats(self, gamePacket: GameTickPacket):
        """
        Update and track stats based on the game packet
        """
        # keep track of demos
        for i in range(len(self._in_demo_state)):
            cur_player = gamePacket.game_cars[i]
            if self._in_demo_state[i]:  # we will toggle this if we have respawned
                self._in_demo_state[i] = cur_player.is_demolished
            elif cur_player.is_demolished:
                print("SOMEONE GOT DEMO'd")
                self._in_demo_state[i] = True
                if i == self._human_player_index:
                    self.stats["recievedDemos"] += 1
                elif i >= self._challenge["humanTeamSize"]:
                    # its an opponent bot
                    self.stats["opponentRecievedDemos"] += 1

        touch = gamePacket.game_ball.latest_touch
        team = touch.team
        self._last_touch_by_team[team] = touch

        for i in range(2):  # iterate of [{team_index, score}]
            team_index = gamePacket.teams[i].team_index
            new_score = gamePacket.teams[i].score
            if new_score != self._last_score_by_team[team_index]:
                self._last_score_by_team[team_index] = new_score

                if team_index == self._human_team:
                    last_touch_player = self._last_touch_by_team[
                        team_index
                    ].player_index
                    if last_touch_player == self._human_player_index:
                        self.stats["humanGoalsScored"] += 1
                        print("humanGoalsScored")


def wait_till_cars_spawned(
    setup_manager: SetupManager, expected_player_count: int
) -> GameTickPacket:
    packet = GameTickPacket()
    setup_manager.game_interface.fresh_live_data_packet(packet, 1000, WITNESS_ID)
    waiting_start = time.monotonic()
    while packet.num_cars != expected_player_count and time.monotonic() - waiting_start < 5:
        print("Game started but no cars are in the packets")
        time.sleep(0.5)
        setup_manager.game_interface.fresh_live_data_packet(packet, 1000, WITNESS_ID)

    return packet


def manage_game_state(
    challenge: dict, upgrades: dict, setup_manager: SetupManager
) -> Tuple[bool, dict]:
    """
    Continuously track the game and adjust state to respect challenge rules and
    upgrades.
    At the end of the game, calculate results and the challenge completion
    and return that
    """
    early_failure = False, {}

    expected_player_count = challenge["humanTeamSize"] + len(challenge["opponentBots"])
    # Wait for everything to be initialized
    packet = wait_till_cars_spawned(setup_manager, expected_player_count)

    if packet.num_cars == 0:
        print("The game was initialized with no cars")
        return early_failure

    tick_rate = 120
    results = None
    max_boost = 0
    if "boost-100" in upgrades:
        max_boost = 100
    elif "boost-33" in upgrades:
        max_boost = 33

    half_field = challenge.get("limitations", []).count("half-field") > 0

    stats_tracker = ManualStatsTracker(challenge)
    last_boost_bump_time = time.monotonic()
    while True:
        try:
            eel.sleep(0)  # yield to allow other gui threads to operate.
            packet = GameTickPacket()
            setup_manager.game_interface.fresh_live_data_packet(
                packet, 1000, WITNESS_ID
            )

            if packet.num_cars == 0:
                # User seems to have ended the match
                print("User ended the match")
                return early_failure

            stats_tracker.updateStats(packet)
            results = packet_to_game_results(packet)

            if has_user_perma_failed(challenge, stats_tracker.stats):
                time.sleep(1)
                setup_failure_freeplay(setup_manager, "You failed the challenge!")
                return early_failure

            if end_by_mercy(challenge, stats_tracker.stats, results):
                time.sleep(3)
                setup_failure_freeplay(setup_manager, "Challenge completed by mercy rule!", "green")
                return True, results

            human_info = packet.game_cars[0]
            game_state = GameState()
            human_desired_state = CarState()
            game_state.cars = {0: human_desired_state}

            changed = False
            # adjust boost
            if human_info.boost > max_boost and not half_field:
                # Adjust boost, unless in heatseeker mode
                human_desired_state.boost_amount = max_boost
                changed = True

            if "boost-recharge" in upgrades:
                # increase boost at 10% per second
                now = time.monotonic()
                if human_info.boost < max_boost and (now - last_boost_bump_time > 0.1):
                    changed = True
                    last_boost_bump_time = now
                    human_desired_state.boost_amount = min(human_info.boost + 1, max_boost)


            if changed:
                setup_manager.game_interface.set_game_state(game_state)

            if packet.game_info.is_match_ended:
                break

        except KeyError:
            traceback.print_exc()
            # it means that the game was interrupted by the user
            print("Looks like the game is in a bad state")
            setup_failure_freeplay(setup_manager, "The game was interrupted.")
            return early_failure

    return calculate_completion(challenge, stats_tracker.stats, results), results




def run_challenge(
    match_config: MatchConfig, challenge: dict, upgrades: dict, launcher_pref: RocketLeagueLauncherPreference
) -> Tuple[bool, dict]:
    """Launch the game and keep track of the state"""
    setup_manager = get_fresh_setup_manager()
    setup_match(setup_manager, match_config, launcher_pref)

    setup_manager.game_interface.renderer.clear_screen(RENDERING_GROUP)
    game_results = None
    try:
        game_results = manage_game_state(challenge, upgrades, setup_manager)
    except:
        # no matter what happens we gotta continue
        traceback.print_exc()
        print("Something failed with the game. Will proceed with shutdown")
        # need to make failure apparent to user
        setup_failure_freeplay(setup_manager, "The game failed to continue")
        return False, {}

    setup_manager.shut_down()

    return game_results


def configure_challenge(challenge: dict, saveState, human_picks: List[int], all_bots):
    player_configs = make_player_configs(
        challenge, human_picks, saveState.team_info, saveState.teammates, all_bots
    )
    match_config = make_match_config(challenge, saveState.upgrades, player_configs)

    return match_config


"""
Manages the story
"""
from datetime import datetime
import json
from os import path

import eel
from PyQt5.QtCore import QSettings

from rlbot_gui.persistence.settings import load_launcher_settings, launcher_preferences_from_map
from rlbot_gui.story.story_challenge_setup import run_challenge, configure_challenge
from rlbot_gui.story.load_story_descriptions import (
    get_bots_configs,
    get_cities,
    get_story_settings,
    get_challenges_by_id
)


CURRENT_STATE = None

##### EEL -- these are the hooks exposed to JS
@eel.expose
def story_story_test():
    print("In story_story_test()")
    # easy way to trigger python code


@eel.expose
def get_cities_json(story_id):
    return get_cities(story_id)


@eel.expose
def get_bots_json(story_id):
    return get_bots_configs(story_id)

@eel.expose
def get_story_settings_json(story_id):
    return get_story_settings(story_id)


@eel.expose
def story_load_save():
    """Loads a previous save if available."""
    global CURRENT_STATE
    settings = QSettings("rlbotgui", "story_save")
    state = settings.value("save")
    if state:
        print(f"Save state: {state}")
        CURRENT_STATE = StoryState.from_dict(state)
        # default values should get added if missing
        state = CURRENT_STATE.__dict__
    return state


@eel.expose
def story_new_save(player_settings, story_settings):
    global CURRENT_STATE
    CURRENT_STATE = StoryState.new(player_settings, story_settings)
    return story_save_state()


@eel.expose
def story_delete_save():
    global CURRENT_STATE
    CURRENT_STATE = None
    QSettings("rlbotgui", "story_save").remove("save")


@eel.expose
def story_save_state():
    settings = QSettings("rlbotgui", "story_save")
    serialized = CURRENT_STATE.__dict__
    settings.setValue("save", serialized)
    return serialized


@eel.expose
def story_save_fake_state(state):
    """Only use for debugging.
    Normally state should just flow from python to JS"""
    global CURRENT_STATE
    settings = QSettings("rlbotgui", "story_save")
    CURRENT_STATE = CURRENT_STATE.from_dict(state)
    settings.setValue("save", state)


@eel.expose
def launch_challenge(challenge_id, pickedTeammates):
    launch_challenge_with_config(challenge_id, pickedTeammates)


@eel.expose
def purchase_upgrade(id, current_currency, cost):
    CURRENT_STATE.add_purchase(id, current_currency, cost)
    flush_save_state()


@eel.expose
def recruit(id, current_currency):
    CURRENT_STATE.add_recruit(id, current_currency)
    flush_save_state()


##### Reverse eel's


def flush_save_state():
    serialized = story_save_state()
    eel.loadUpdatedSaveState(serialized)


#################################


class StoryState:
    """Represents users game state"""

    def __init__(self):
        self.version = 1
        self.story_config = "default" # can be dict for custom config
        self.team_info = {"name": "", "color_secondary": ""}
        self.teammates = []
        self.challenges_attempts = {}  # many entries per challenge
        self.challenges_completed = {}  # one entry per challenge

        self.upgrades = {"currency": 0}

    def add_purchase(self, id, current_currency, cost):
        """The only validation we do is to make sure current_currency is correct.
        This is NOT a security mechanism, this is a bug prevention mechanism to
        avoid accidental double clicks.
        """
        if self.upgrades["currency"] == current_currency:
            self.upgrades[id] = True
            self.upgrades["currency"] -= cost

    def add_recruit(self, id, current_currency):
        """The only validation we do is to make sure current_currency is correct.
        This is NOT a security mechanism, this is a bug prevention mechanism to
        avoid accidental double clicks.
        """
        if self.upgrades["currency"] == current_currency:
            self.teammates.append(id)
            self.upgrades["currency"] -= 1

    def add_match_result(
        self, challenge_id: str, challenge_completed: bool, game_results
    ):
        """game_results should be the output of packet_to_game_results.
        You have to call it anyways to figure out if the player
        completed the challenge so that's why we don't call it again here.
        """
        if challenge_id not in self.challenges_attempts:
            # no defaultdict because we serialize the data
            self.challenges_attempts[challenge_id] = []

        self.challenges_attempts[challenge_id].append(
            {"game_results": game_results, "challenge_completed": challenge_completed}
        )

        if challenge_completed:
            index = len(self.challenges_attempts[challenge_id]) - 1
            self.challenges_completed[challenge_id] = index
            self.upgrades["currency"] += 2

    @staticmethod
    def new(player_settings, story_settings):
        s = StoryState()

        name = player_settings["name"]
        color_secondary = player_settings["color"]
        s.team_info = {"name": name, "color_secondary": color_secondary}

        story_id = story_settings["story_id"]
        custom_config = story_settings["custom_config"]
        use_custom_maps = story_settings["use_custom_maps"]
        if story_id == 'custom':
            s.story_config = custom_config
        else:
            s.story_config = story_id
            if use_custom_maps:
                s.story_config = f"{story_id}-with-cmaps"
        return s

    @staticmethod
    def from_dict(source):
        """No validation done here."""
        s = StoryState()
        s.__dict__.update(source)
        return s


def launch_challenge_with_config(challenge_id, pickedTeammates):
    print(f"In launch_challenge {challenge_id}")

    story_id = CURRENT_STATE.story_config
    challenge = get_challenges_by_id(story_id)[challenge_id]
    all_bots = get_bots_configs(story_id)


    match_config = configure_challenge(challenge, CURRENT_STATE, pickedTeammates, all_bots)
    launcher_settings_map = load_launcher_settings()
    launcher_prefs = launcher_preferences_from_map(launcher_settings_map)
    completed, results = run_challenge(match_config, challenge, CURRENT_STATE.upgrades, launcher_prefs)
    CURRENT_STATE.add_match_result(challenge_id, completed, results)

    flush_save_state()




def convert_packet_to_dict(ctypes_packet):
    """
    Here we're selectively converting parts of the game tick packet into dict format.
    We're skipping over parts that are currently not needed by the GUI, e.g. boost pad state.
    """
    dict_version = {}
    dict_version['game_ball'] = getdict(ctypes_packet.game_ball)
    dict_version['game_info'] = getdict(ctypes_packet.game_info)
    cars = []
    for i in range(ctypes_packet.num_cars):
        cars.append(getdict(ctypes_packet.game_cars[i]))
    dict_version['game_cars'] = cars
    return dict_version


def getdict(struct):
    """
    This will convert a ctypes struct into a python dict. After that, it can be converted easily to json.
    Taken from https://stackoverflow.com/a/34301571/280852
    """
    result = {}

    def get_value(value):
        if (type(value) not in [int, float, bool]) and not bool(value):
            # it's a null pointer
            value = None
        elif hasattr(value, "_length_") and hasattr(value, "_type_"):
            # Probably an array
            # print value
            value = get_array(value)
        elif hasattr(value, "_fields_"):
            # Probably another struct
            value = getdict(value)
        return value

    def get_array(array):
        ar = []
        for value in array:
            value = get_value(value)
            ar.append(value)
        return ar

    for f in struct._fields_:
        field = f[0]
        value = getattr(struct, field)
        # if the type is not a primitive and it evaluates to False ...
        value = get_value(value)
        result[field] = value
    return result



"""
Helpers to load and set up matches in custom maps
"""

from contextlib import contextmanager
from datetime import datetime
from os import path

from typing import List, Optional

import glob
import shutil
import os

from rlbot.setup_manager import (
    SetupManager,
    RocketLeagueLauncherPreference,
    try_get_steam_executable_path,
)
from rlbot.gamelaunch.epic_launch import locate_epic_games_launcher_rocket_league_binary
from rlbot.utils import logging_utils

from rlbot_gui.persistence.settings import load_settings, BOT_FOLDER_SETTINGS_KEY

CUSTOM_MAP_TARGET = {"filename": "Labs_Utopia_P.upk", "game_map": "UtopiaRetro"}

logger = logging_utils.get_logger("custom_maps")


@contextmanager
def prepare_custom_map(custom_map_file: str, rl_directory: str):
    """
    Provides a context manager. It will swap out the custom_map_file
    for an existing map in RL and it will return the `game_map`
    name that should be used in a MatchConfig.

    Once the context is left, the original map is replaced back.
    The context should be left as soon as the match has started
    """

    # check if there metadata for the custom file
    expected_config_name = "_" + path.basename(custom_map_file)[:-4] + ".cfg"
    config_path = path.join(path.dirname(custom_map_file), expected_config_name)
    additional_info = {
        "original_path": custom_map_file,
    }
    if path.exists(config_path):
        additional_info["config_path"] = config_path


    real_map_file = path.join(rl_directory, CUSTOM_MAP_TARGET["filename"])
    timestamp = datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
    temp_filename = real_map_file + "." + timestamp

    shutil.copy2(real_map_file, temp_filename)
    logger.info("Copied real map to %s", temp_filename)
    shutil.copy2(custom_map_file, real_map_file)
    logger.info("Copied custom map from %s", custom_map_file)

    try:
        yield CUSTOM_MAP_TARGET["game_map"], additional_info
    finally:
        os.replace(temp_filename, real_map_file)
        logger.info("Reverted real map to %s", real_map_file)


def convert_custom_map_to_path(custom_map: str) -> Optional[str]:
    """
    Search through user's selected folders to find custom_map
    Return none if not found, full path if found
    """
    custom_map_file = None
    folders = get_search_folders()
    for folder in folders:
        scan_query = path.join(glob.escape(folder), "**", custom_map)
        for match in glob.iglob(scan_query, recursive=True):
            custom_map_file = match

    if not custom_map_file:
        logger.warning("%s - map doesn't exist", custom_map)

    return custom_map_file


def find_all_custom_maps() -> List[str]:
    """
    Ignores maps starting with _
    """
    folders = get_search_folders()
    maps = []
    for folder in folders:
        scan_query = path.join(glob.escape(folder), "**", "*.u[pd]k")
        for match in glob.iglob(scan_query, recursive=True):
            basename = path.basename(match)
            if basename.startswith("_"):
                continue
            maps.append(basename)
    return maps


def get_search_folders() -> List[str]:
    """Get all folders to search for maps"""
    bot_folders_setting = load_settings().value(BOT_FOLDER_SETTINGS_KEY, type=dict)
    folders = {}
    if "folders" in bot_folders_setting:
        folders = bot_folders_setting["folders"]
    return [k for k, v in folders.items() if v['visible']]


def identify_map_directory(launcher_pref: RocketLeagueLauncherPreference):
    """Find RocketLeague map directory"""
    final_path = None
    if launcher_pref.preferred_launcher == RocketLeagueLauncherPreference.STEAM:
        steam = try_get_steam_executable_path()
        suffix = r"steamapps\common\rocketleague\TAGame\CookedPCConsole"
        if not steam:
            return None

        # TODO: Steam can install RL on a different disk. Need to
        # read libraryfolders.vdf to detect this situation
        # It's a human-readable but custom format so not trivial to parse

        final_path = path.join(path.dirname(steam), suffix)
    else:
        rl_executable = locate_epic_games_launcher_rocket_league_binary()
        suffix = r"TAGame\CookedPCConsole"
        if not rl_executable:
            return None

        # Binaries/Win64/ is what we want to strip off
        final_path = path.join(path.dirname(rl_executable), "..", "..", suffix)

    if not path.exists(final_path):
        logger.warning("%s - directory doesn't exist", final_path)
        return None
    return final_path



import fileinput
import os
import random
import re
import string
import sys
import tempfile
from pathlib import Path
from shutil import move

from rlbot.parsing.directory_scanner import scan_directory_for_bot_configs

from rlbot_gui.bot_management.downloader import download_and_extract_zip


def convert_to_filename(text):
    """
    Normalizes string, converts to lowercase, removes non-alphanumeric characters,
    and converts spaces to underscores.
    """
    import unicodedata
    normalized = unicodedata.normalize('NFKD', text).encode('ascii', 'ignore').decode()
    valid_chars = f'-_.() {string.ascii_letters}{string.digits}'
    filename = ''.join(c for c in normalized if c in valid_chars)
    filename = filename.replace(' ', '_')  # Replace spaces with underscores
    return filename


def safe_move(src, dst):
    """ https://bugs.python.org/issue32689 """
    move(str(src), str(dst))


def replace_all(file, regex, replacement):
    for line in fileinput.input(file, inplace=1):
        updated = re.sub(regex, replacement, line)
        sys.stdout.write(updated)


def bootstrap_python_bot(bot_name, directory):
    sanitized_name = convert_to_filename(bot_name)
    bot_directory = Path(directory or '.')
    top_dir = bot_directory / sanitized_name
    if os.path.exists(top_dir):
        raise FileExistsError(f'There is already a bot named {sanitized_name}, please choose a different name!')

    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdir = Path(tmpdirname)
        print('created temporary directory', tmpdir)

        download_and_extract_zip(
            download_url='https://github.com/RLBot/RLBotPythonExample/archive/master.zip',
            local_folder_path=tmpdir)

        safe_move(tmpdir / 'RLBotPythonExample-master', top_dir)

    bundle = scan_directory_for_bot_configs(top_dir).pop()
    config_file = bundle.config_path
    python_file = bundle.python_file

    replace_all(config_file, r'name = .*$', 'name = ' + bot_name)

    # This is intended to open the example python file in the default system editor for .py files.
    # Hopefully this will be VS Code or notepad++ or something. If it gets executed as a python script, no harm done.
    os.startfile(python_file)

    return config_file


def bootstrap_scratch_bot(bot_name, directory):
    sanitized_name = convert_to_filename(bot_name)
    bot_directory = Path(directory or '.')
    top_dir = bot_directory / sanitized_name
    if os.path.exists(top_dir):
        raise FileExistsError(f'There is already a bot named {sanitized_name}, please choose a different name!')

    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdir = Path(tmpdirname)
        print('created temporary directory', tmpdir)

        download_and_extract_zip(
            download_url='https://github.com/RLBot/RLBotScratchInterface/archive/gui-friendly.zip',
            local_folder_path=tmpdir)

        safe_move(tmpdir / 'RLBotScratchInterface-gui-friendly', top_dir)

    # Choose appropriate file names based on the bot name
    code_dir = top_dir / sanitized_name
    sb3_filename = f'{sanitized_name}.sb3'
    sb3_file = code_dir / sb3_filename
    config_filename = f'{sanitized_name}.cfg'
    config_file = code_dir / config_filename

    replace_all(top_dir / 'rlbot.cfg', r'(participant_config_\d = ).*$',
                r'\1' + os.path.join(sanitized_name, config_filename).replace('\\', '\\\\'))

    # We're assuming that the file structure / names in RLBotScratchInterface will not change.
    # Semi-safe assumption because we're looking at a gui-specific git branch which ought to be stable.
    safe_move(top_dir / 'scratch_bot', code_dir)
    safe_move(code_dir / 'my_scratch_bot.sb3', sb3_file)
    safe_move(code_dir / 'my_scratch_bot.cfg', config_file)

    replace_all(config_file, r'name = .*$', 'name = ' + bot_name)
    replace_all(config_file, r'sb3file = .*$', 'sb3file = ' + sb3_filename)
    replace_all(config_file, r'port = .*$', 'port = ' + str(random.randint(20000, 65000)))

    return config_file


def bootstrap_python_hivemind(hive_name, directory):
    sanitized_name = convert_to_filename(hive_name)
    bot_directory = Path(directory or '.')
    top_dir = bot_directory / sanitized_name
    if os.path.exists(top_dir):
        raise FileExistsError(f'There is already a bot named {sanitized_name}, please choose a different name!')

    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdir = Path(tmpdirname)
        print('created temporary directory', tmpdir)

        download_and_extract_zip(
            download_url='https://github.com/RLBot/RLBotPythonHivemindExample/archive/master.zip',
            local_folder_path=tmpdir)

        safe_move(tmpdir / 'RLBotPythonHivemindExample-master', top_dir)


    config_file = top_dir / 'config.cfg'
    drone_file = top_dir / 'src' / 'drone.py'
    hive_file = top_dir / 'src' / 'hive.py'

    replace_all(config_file, r'name = .*$', f'name = {hive_name}')
    replace_all(drone_file, r'hive_name = .*$', f'hive_name = "{hive_name} Hivemind"')
    replace_all(drone_file, r'hive_key = .*$', f'hive_key = "{random.randint(100000, 999999) + hash(hive_name)}"')
    replace_all(hive_file, r'class .*\(PythonHivemind\)', f'class {hive_name}Hivemind(PythonHivemind)')

    # This is intended to open the example python file in the default system editor for .py files.
    # Hopefully this will be VS Code or notepad++ or something. If it gets executed as a python script, no harm done.
    os.startfile(hive_file)

    return config_file
